/*
   INDI Developers Manual
   Tutorial #3
   
   "Binary data transfer"
   
   In this tutorial, we establish a data channel with the client to send a FITS file. 
   
   Refer to README, which contains instruction on how to build this driver, and use it 
   with an INDI-compatible client.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/* INDI Core headers */
#include "indidevapi.h"
#include "indicom.h"
#include "eventloop.h"

/* We need to use the minizlo compression library to compress data before we send it */
#include "../lzo/minilzo.h"

#define mydev         		"Device with Data Transfer"
#define COMM_GROUP	"Main Control"
#define DATA_GROUP      "Data Channel"

#define FRAME_ILEN	64
#define PIPEBUFSIZ		8192

void ISInit(void);
void connectDevice(void);
void uploadFile(char * filename);
int   checkPowerS(ISwitchVectorProperty *sp);

/* Functions used to establish and maintain the binary channel */
void initDataChannel(void);
void waitForData(int rp, int wp);
void updateDataChannel(void *p);
int    findPort(void);

/* Definitions for minizlo */
#define HEAP_ALLOC(var,size) \
	lzo_align_t __LZO_MMODEL var [ ((size) + (sizeof(lzo_align_t) - 1)) / sizeof(lzo_align_t) ]

static HEAP_ALLOC(wrkmem,LZO1X_1_MEM_COMPRESS);

typedef struct {
 int rp;				/* Read pipe */
 int wp;				/* Write pipe */
 int streamReady;		/* Can we write to data stream? */
} client_t;

client_t *INDIClients;			/* INDI clients */
static int DataPort;			/* Data Port */
static int nclients;			/* # of clients using the binary stream */
static int dataChannelTimerID;	/* Timer ID */
static int disableUpdate;		/* disable update for binary channel */

/*INDI controls */
/* Connect/Disconnect */
static ISwitch PowerS[]          			= {{"CONNECT" , "Connect" , ISS_OFF, 0, 0},{"DISCONNECT", "Disconnect", ISS_ON, 0, 0}};
static ISwitchVectorProperty PowerSP	= { mydev, "CONNECTION" , "Connection", COMM_GROUP, IP_RW, ISR_1OFMANY, 60, IPS_IDLE, PowerS, NARRAY(PowerS), 0, 0};

/* Send data switch */
static ISwitch SendDataS[]  			= { {"FITS", "", ISS_OFF, 0, 0 } };
static ISwitchVectorProperty SendDataSP 	= { mydev, "Download", "", COMM_GROUP, IP_RW, ISR_1OFMANY, 0, IPS_IDLE, SendDataS, NARRAY(SendDataS), 0, 0};
  
/* Data channel */
static INumber DataChannelN[]		= {{"CHANNEL", "Channel", "%0.f", 1024., 20000., 1., 0., 0, 0, 0}};
static INumberVectorProperty DataChannelNP={ mydev, "DATA_CHANNEL", "Data Channel", DATA_GROUP, IP_RO, 0, IPS_IDLE, DataChannelN, NARRAY(DataChannelN), 0, 0};

/* send client definitions of all properties */
void ISInit()
{
  static int isInit=0;

 if (isInit)
  return;
 
 dataChannelTimerID = -1;
 INDIClients 		= NULL;
 nclients    		= 0;
 
 /* Initlaize the minilzo library */
 if (lzo_init() != LZO_E_OK)
   IDLog("lzo_init() failed !!!\n");
      
 isInit = 1;

}

void ISGetProperties (const char *dev)
{ 

   ISInit();
  
  if (dev && strcmp (mydev, dev))
    return;
    
  /* COMM_GROUP */
  IDDefSwitch(&PowerSP, NULL);
  IDDefSwitch(&SendDataSP, NULL);
 
  /* Data Group */
  IDDefNumber(&DataChannelNP, NULL);

  /* Our driver should be able to communicate to any number of clients*/
  INDIClients = INDIClients ? (client_t *) realloc(INDIClients, (nclients+1) * sizeof(client_t)) :
                              (client_t *) malloc (sizeof(client_t));
    
  if (INDIClients)
  {
        INDIClients[nclients].streamReady = 0;
	INDIClients[nclients].rp	 	   = 0;
	INDIClients[nclients].wp          	   = 0;
  	DataPort = findPort();
        if (DataPort > 0)
 		initDataChannel();
  }
  
}
  
void ISNewSwitch (const char *dev, const char *name, ISState *states, char *names[], int n)
{
	/* ignore if not ours */
	if (dev && strcmp (dev, mydev))
	    return;
	    
	ISInit();
     
     /* Connection */
     if (!strcmp (name, PowerSP.name))
     {
          IUResetSwitches(&PowerSP);
	  IUUpdateSwitches(&PowerSP, states, names, n);
   	  connectDevice();
	  return;
     }
    
  
     /* Send data */
    if (!strcmp (SendDataSP.name, name))
    {
       
       /* If power is off, return */
       if (checkPowerS(&SendDataSP))
         return;
    
	IUResetSwitches(&SendDataSP);
	SendDataSP.s = IPS_OK;
	IDSetSwitch(&SendDataSP, "Uploading a fits file...");
	uploadFile("Abell426.fts");
	return;
     }
     
}

void ISNewText (const char *dev, const char *name, char *texts[], char *names[], int n)
{
	return;
	
}


void ISNewNumber (const char *dev, const char *name, double values[], char *names[], int n)
{
        return;	
}

/* If the connection is on, return 0, otherwise, send an error message to the client and return -1 */
int checkPowerS(ISwitchVectorProperty *sp)
{
  if (PowerSP.s != IPS_OK)
  {
    if (!strcmp(sp->label, ""))
        IDMessage (mydev, "Cannot change property %s while the device is offline.", sp->name);
    else
    	IDMessage (mydev, "Cannot change property %s while the device is offline.", sp->label);
	
    sp->s = IPS_IDLE;
    IDSetSwitch(sp, NULL);
    return -1;
  }

  return 0;
}

void connectDevice()
{
   
  switch (PowerS[0].s)
  {
     case ISS_ON:
     
       PowerSP. s = IPS_OK;
       IDSetSwitch(&PowerSP, "Device is connected.");
	
      if (dataChannelTimerID == -1)
      {
        disableUpdate = 0;
        updateDataChannel(NULL);
      }
      
      break;
      
    case ISS_OFF:
      
      PowerSP.s = IPS_IDLE;
      disableUpdate = 1;
      IDSetSwitch(&PowerSP, "Device is disconnected.");
      
      break;
     }
}

/* Upload file to client */
void uploadFile(char * filename)
{
   FILE * fitsFile;
   char frameSize[FRAME_ILEN];
   unsigned char *fitsData, *compressedFrame;
   int n=0, r=0;
   unsigned int totalBytes, compressedBytes = 0, frameLen =0 , nr = 0, i=0, j = 0;
   struct stat stat_p; 
 
   /* Get file size */
   if ( -1 ==  stat (filename, &stat_p))
   { 
     IDLog(" Error occoured attempting to stat %s\n", filename); 
     return; 
   }
   
   totalBytes = stat_p.st_size;
   fitsData = (unsigned char *) malloc (sizeof(unsigned char) * totalBytes);

   /* Open file */
   fitsFile = fopen(filename, "r");
   
   if (fitsFile == NULL)
    return;
   
   /* Read file from disk */ 
   for (i=0; i < totalBytes; i+= nr)
   {
      nr = fread(fitsData + i, 1, totalBytes - i, fitsFile);
     
     if (nr <= 0)
     {
        IDLog("Error reading temporary FITS file.\n");
        return;
     }
   }
   
   /* Compress it */
   compressedFrame = (unsigned char *) malloc (sizeof(unsigned char) * totalBytes + totalBytes / 64 + 16 + 3);
   r = lzo1x_1_compress(fitsData, totalBytes, compressedFrame, &compressedBytes, wrkmem);
   if (r != LZO_E_OK)
   {
 	/* this should NEVER happen */
 	IDLog("internal error - compression failed: %d\n", r);
	return;
   }
   
   /* Format the frame header */
   snprintf(frameSize, FRAME_ILEN, "<Data type='FITS' size='%d' compsize='%d' />", totalBytes, compressedBytes);
   frameLen = strlen(frameSize); 
   r = 0;
   
     /* Send to all clients */
     for (i=0; i < nclients; i++)
     {
  	if (INDIClients[i].streamReady)
        {
	      /* send frame header to client */
	     for (j=0; j < frameLen; j+=r)
	      r = write(INDIClients[i].wp, frameSize + j , frameLen - j);
	      
	     /* Send the actual binary data */
             for (nr=0; nr < compressedBytes; nr+=n)
   	     {
	  	n = write(INDIClients[i].wp, compressedFrame + nr, compressedBytes - nr );
		if (n <= 0)
		{
			if (nr <= 0)
    			{
      				IDLog("Stream error: %s\n", strerror(errno));
      				return;
    			}
		        break;
		}
		
             }
        }
      } 
      
    
   free (fitsData);
   free (compressedFrame);
}


void initDataChannel ()
{
	int pid;
	int rp[2], wp[2];

	/* new pipes */
	if (pipe (rp) < 0) {
	    fprintf (stderr, "%s: read pipe: %s\n", mydev, strerror(errno));
	    exit(1); 
	}
	if (pipe (wp) < 0) {
	    fprintf (stderr, "%s: write pipe: %s\n", mydev, strerror(errno));
	    exit(1);
	}

	/* new process */
	pid = fork();
	if (pid < 0) {
	    fprintf (stderr, "%s: fork: %s\n", mydev, strerror(errno));
	    exit(1);
	}
	if (pid == 0) {
	    /* child: listen to driver */
	    close(wp[1]);
	    close(rp[0]);
	    
	    waitForData(wp[0], rp[1]);
	    
	    return;
	}

	/* Assign read and write pipes */
	close (wp[0]);
	close (rp[1]);
	INDIClients[nclients].rp  = rp[0];
	INDIClients[nclients].wp = wp[1];
	
	nclients++;
	
	IDLog("Going to listen on fd %d, and write to fd %d\n", rp[0], wp[1]);
	updateDataChannel(NULL);
}

void updateDataChannel(void *p)
{
  char buffer[1];
  fd_set rs;
  int nr, i=0;
  p=p;
  struct timeval tv;
  tv.tv_sec  = 0;
  tv.tv_usec = 0;
  
  if (disableUpdate)
  {
   dataChannelTimerID = -1;
   return;
  }
  
  for (i=0; i < nclients; i++)
  {
        if (INDIClients[i].rp < 0)
	 continue;
	 
  	FD_ZERO(&rs);
  	FD_SET(INDIClients[i].rp, &rs);
  
  	nr = select(INDIClients[i].rp+1, &rs, NULL, NULL, &tv);
  
  	if (nr <= 0)
	 continue;

  	nr = read(INDIClients[i].rp, buffer, 1);
	buffer[1]='\0';
  	if (nr > 0 && atoi(buffer) == 0)
	{
	        IDLog("Client %d is ready to receive stream\n", i);
		INDIClients[i].streamReady = 1;
	}
	else
	{
		INDIClients[i].streamReady = 0;
		INDIClients[i].rp          = -1;
		IDLog("Lost connection with client %d\n", i);
	}
   }
  
   dataChannelTimerID = addTimer(1000, updateDataChannel, NULL);    
}

/* A simple function to find a free port for our binary channel */
int findPort()
{
  struct sockaddr_in serv_socket;
  int sfd;
  int port = 8000;
  int i=0;
  
  /* make socket endpoint */
  if ((sfd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	    fprintf (stderr, "%s: socket: %s", mydev, strerror(errno));
	    exit(1);
  }
	
  /* bind to given port for any IP address */
  memset (&serv_socket, 0, sizeof(serv_socket));
  serv_socket.sin_family = AF_INET;
  serv_socket.sin_addr.s_addr = htonl (INADDR_ANY);
  
  for (i=0; i < 100; i++)
  {
  	serv_socket.sin_port = htons ((unsigned short)port);
	if (bind(sfd,(struct sockaddr*)&serv_socket,sizeof(serv_socket)) < 0)
	{
	  	fprintf (stderr, "%s: bind: %s\n", mydev, strerror(errno));
		port +=5;
	}
	else break;
  }
  
  close(sfd);
  if (i == 100) return -1;
  
  return port;
	
}
	
/* This function will be active in the forked process. It will wait for new data as it arrives from the driver and send them to the client */
void waitForData(int rp, int wp)
{
        struct sockaddr_in serv_socket;
	struct sockaddr_in cli_socket;
	int cli_len;
	int sfd, lsocket;
	int reuse = 1;
        fd_set rs;
	int maxfd;
	int i, s, nr, n;
	unsigned char buffer[PIPEBUFSIZ];
	char dummy[8];
	
	/* make socket endpoint */
	if ((sfd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	    fprintf (stderr, "%s: socket: %s", mydev, strerror(errno));
	    exit(1);
	}
	
	/* bind to given port for any IP address */
	memset (&serv_socket, 0, sizeof(serv_socket));
	serv_socket.sin_family = AF_INET;
	serv_socket.sin_addr.s_addr = htonl (INADDR_ANY);
	
	serv_socket.sin_port = htons ((unsigned short)DataPort);
	if (setsockopt(sfd,SOL_SOCKET,SO_REUSEADDR,&reuse,sizeof(reuse)) < 0)
	{
		fprintf (stderr, "%s: setsockopt: %s", mydev, strerror(errno));
	    	exit(1);
	}
	
	if (bind(sfd,(struct sockaddr*)&serv_socket,sizeof(serv_socket)) < 0)
	{
	    	fprintf (stderr, "%s: bind: %s\n", mydev, strerror(errno));
		exit(1);
	}
	
        /* Tell client we're ready to connect */
	DataChannelN[0].value = DataPort;
        DataChannelNP.s = IPS_OK;
        IDSetNumber(&DataChannelNP, NULL);
	
	/* willing to accept connections with a backlog of 1 pending */
	if (listen (sfd, 1) < 0)
	{
	    fprintf (stderr, "%s: listen: %s", mydev, strerror(errno));
	    exit(1);
	}

	/* ok */
	lsocket = sfd;
        fprintf (stderr, "%s: listening to data port %d on fd %d\n", mydev ,DataPort,sfd);

	/* start with public contact point */
	FD_ZERO(&rs);
	FD_SET(lsocket, &rs);
	maxfd = lsocket;

	/* wait for action */
	s = select (maxfd+1, &rs, NULL, NULL, NULL);
	if (s < 0) 
	{
	    fprintf (stderr, "%s: select: %s\n", mydev, strerror(errno));
	    exit(1);
	}
 
	/* get a private connection to new client */
	cli_len = sizeof(cli_socket);
	s = accept (lsocket, (struct sockaddr *)&cli_socket, (socklen_t *)&cli_len);
	if(s < 0) 
	{
	    fprintf (stderr, "%s: accept: %s", mydev, strerror(errno));
	    exit (1);
	}
	
	/* Tell the driver that this client established a connection to the data channel and is ready
	   to receive stream data */
	sprintf(dummy, "%d", 0);
	write(wp, dummy, 1);
	
	FD_ZERO(&rs);
	FD_SET(rp, &rs);
	
	while (1)
	{
	        /* Wait for action */
        	i = select(rp+1, &rs, NULL, NULL, NULL);
		if (i < 0)
		{
	  		fprintf (stderr, "%s: select: %s\n", mydev, strerror(errno));
			/* Tell driver the client has disconnected the data channel */
			sprintf(dummy, "%d", -1);
			write(wp, dummy, 1);
	  		exit(1);
		}
		else if (i==0) continue;
		
		/* Read some data */
		nr = read(rp, buffer, PIPEBUFSIZ);
		if (nr < 0) {
		sprintf(dummy, "%d", -1);
		write(wp, dummy, 1);
	    	exit(1);
		}
		if (nr == 0) 
		{
		sprintf(dummy, "%d", -1);
		write(wp, dummy, 1);
		fprintf (stderr, "Data Client %d: EOF\n", s);
	    	exit(1);
		continue;
		}
		
		/* Write to client */
		for (i=0; i < nr; i+=n)
		{
		  n = write(s, buffer + i, nr - i);
		  
		  if (n <= 0)
		  {
		        fprintf(stderr, "nr is %d\n", nr);
			
			if (n < 0)
		    		fprintf(stderr, "%s\n", strerror(errno));
			else
		    	        fprintf(stderr, "Short write of FITS pixels.\n");
			
		       sprintf(dummy, "%d", -1);
		       write(wp, dummy, 1);
		       exit(1);
		  }
		  
		}
		
	}

}
